# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service entry point."""
import logging
import os
import traceback
import uuid

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import Flask, request
from flask_apispec import FlaskApiSpec
from flask_swagger_ui import get_swaggerui_blueprint

from renku.service.cache import cache
from renku.service.config import API_SPEC_URL, API_VERSION, CACHE_DIR, \
    OPENAPI_VERSION, SERVICE_NAME, SWAGGER_URL
from renku.service.logger import service_log
from renku.service.utils.json_encoder import SvcJSONEncoder
from renku.service.views.cache import CACHE_BLUEPRINT_TAG, cache_blueprint, \
    list_projects_view, list_uploaded_files_view, project_clone_view, \
    upload_file_view
from renku.service.views.datasets import DATASET_BLUEPRINT_TAG, \
    add_file_to_dataset_view, create_dataset_view, dataset_blueprint, \
    edit_dataset_view, import_dataset_view, list_dataset_files_view, \
    list_datasets_view
from renku.service.views.jobs import JOBS_BLUEPRINT_TAG, jobs_blueprint, \
    list_jobs
from renku.service.views.templates import TEMPLATES_BLUEPRINT_TAG, \
    read_manifest_from_template, templates_blueprint

logging.basicConfig(level=os.getenv('SERVICE_LOG_LEVEL', 'WARNING'))


def create_app():
    """Creates a Flask app with a necessary configuration."""
    app = Flask(__name__)
    app.secret_key = os.getenv('RENKU_SVC_SERVICE_KEY', uuid.uuid4().hex)
    app.json_encoder = SvcJSONEncoder
    app.config['UPLOAD_FOLDER'] = CACHE_DIR

    max_content_size = os.getenv('MAX_CONTENT_LENGTH')
    if max_content_size:
        app.config['MAX_CONTENT_LENGTH'] = max_content_size

    app.config['cache'] = cache

    build_routes(app)

    @app.route('/health')
    def health():
        import renku
        return 'renku repository service version {}\n'.format(
            renku.__version__
        )

    return app


def build_routes(app):
    """Register routes to given app instance."""
    app.config.update({
        'APISPEC_SPEC':
            APISpec(
                title=SERVICE_NAME,
                openapi_version=OPENAPI_VERSION,
                version=API_VERSION,
                plugins=[MarshmallowPlugin()],
            ),
        'APISPEC_SWAGGER_URL': API_SPEC_URL,
    })
    app.register_blueprint(cache_blueprint)
    app.register_blueprint(dataset_blueprint)
    app.register_blueprint(jobs_blueprint)
    app.register_blueprint(templates_blueprint)

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL, API_SPEC_URL, config={'app_name': 'Renku Service'}
    )
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    docs = FlaskApiSpec(app)

    docs.register(list_uploaded_files_view, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(upload_file_view, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(project_clone_view, blueprint=CACHE_BLUEPRINT_TAG)
    docs.register(list_projects_view, blueprint=CACHE_BLUEPRINT_TAG)

    docs.register(list_datasets_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(list_dataset_files_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(add_file_to_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(create_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(import_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)
    docs.register(edit_dataset_view, blueprint=DATASET_BLUEPRINT_TAG)

    docs.register(list_jobs, blueprint=JOBS_BLUEPRINT_TAG)
    docs.register(
        read_manifest_from_template, blueprint=TEMPLATES_BLUEPRINT_TAG
    )


app = create_app()


@app.after_request
def after_request(response):
    """After request handler."""
    service_log.info(
        '{0} {1} {2} {3} {4}'.format(
            request.remote_addr, request.method, request.scheme,
            request.full_path, response.status
        )
    )

    return response


@app.errorhandler(Exception)
def exceptions(e):
    """App exception logger."""
    tb = traceback.format_exc()
    service_log.error(
        '{} {} {} {} 5xx INTERNAL SERVER ERROR\n{}'.format(
            request.remote_addr, request.method, request.scheme,
            request.full_path, tb
        )
    )

    return e.status_code


if __name__ == '__main__':
    app.logger.handlers.extend(service_log.handlers)
    app.run()



# -*- coding: utf-8 -*-
#
# Copyright 2020 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataverse unit tests."""
from renku.core.commands.providers.dataverse import DataverseExporter, \
    _DataverseDeposition


def test_dataverse_exporter_init():
    """Check construction of a dataverse exporter."""
    exporter = DataverseExporter(
        dataset='my-dataset',
        access_token='let-me-in',
    )

    assert exporter
    assert 'my-dataset' == exporter.dataset
    assert 'let-me-in' == exporter.access_token


def test_dataverse_deposition_init():
    """Check construction of dataverse deposition."""
    deposit = _DataverseDeposition(
        access_token='let-me-in',
        server_url='http://sss',
    )

    assert deposit
    assert 'let-me-in' == deposit.access_token
    assert 'http://sss' == deposit.server_url
    assert not deposit.dataset_pid



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service headers serializers."""
from marshmallow import Schema, ValidationError, fields, pre_load
from werkzeug.utils import secure_filename


class UserIdentityHeaders(Schema):
    """User identity schema."""

    user_id = fields.String(required=True, data_key='renku-user-id')
    fullname = fields.String(data_key='renku-user-fullname')
    email = fields.String(data_key='renku-user-email')
    token = fields.String(data_key='authorization')

    def extract_token(self, data):
        """Extract token."""
        value = data.get('authorization', '')
        components = value.split(' ')

        rfc_compliant = value.lower().startswith('bearer')
        rfc_compliant &= len(components) == 2

        if not rfc_compliant:
            raise ValidationError('authorization value contains invalid value')

        return components[-1]

    @pre_load()
    def set_fields(self, data, **kwargs):
        """Set fields for serialization."""
        expected_keys = [field.data_key for field in self.fields.values()]

        data = {
            key.lower(): value
            for key, value in data.items() if key.lower() in expected_keys
        }

        if {'renku-user-id', 'authorization'}.issubset(set(data.keys())):
            data['renku-user-id'] = secure_filename(data['renku-user-id'])
            data['authorization'] = self.extract_token(data)

        return data



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template serializers."""

from marshmallow import Schema, fields, pre_load

from renku.service.config import TEMPLATE_CLONE_DEPTH_DEFAULT
from renku.service.serializers.cache import ProjectCloneContext
from renku.service.serializers.rpc import JsonRPCResponse


class ManifestTemplatesRequest(ProjectCloneContext):
    """Request schema for listing manifest templates."""

    url = fields.String(required=True)
    ref = fields.String(missing='master')
    depth = fields.Integer(missing=TEMPLATE_CLONE_DEPTH_DEFAULT)

    @pre_load()
    def set_git_url(self, data, **kwargs):
        """Set git_url field."""
        data['git_url'] = data['url']

        return data


class ManifestTemplateSchema(Schema):
    """Manifest template schema."""

    description = fields.String(required=True)
    folder = fields.String(required=True)
    name = fields.String(required=True)
    variables = fields.Dict(missing={})


class ManifestTemplatesResponse(Schema):
    """Manifest templates response."""

    templates = fields.List(
        fields.Nested(ManifestTemplateSchema), required=True
    )


class ManifestTemplatesResponseRPC(JsonRPCResponse):
    """RPC schema for listing manifest templates."""

    result = fields.Nested(ManifestTemplatesResponse)



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper utils for handling UUIDs."""
import uuid


def is_uuid(value):
    """Check if value is UUID4.

    Copied from https://stackoverflow.com/questions/19989481/
    """
    try:
        uuid_obj = uuid.UUID(value, version=4)
    except ValueError:
        return False

    return str(uuid_obj) == value



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view tests."""
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from flaky import flaky
from tests.core.commands.test_init import TEMPLATE_ID, TEMPLATE_INDEX, \
    TEMPLATE_REF, TEMPLATE_URL

from renku.core.commands.init import fetch_template, read_template_manifest


@pytest.mark.service
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_read_manifest_from_template(svc_client_with_templates):
    """Check reading manifest template."""
    svc_client, headers, template_params = svc_client_with_templates

    response = svc_client.get(
        '/templates.read_manifest',
        query_string=template_params,  # data=json.dumps(payload)
        headers=headers
    )

    assert response
    assert {'result'} == set(response.json.keys())
    assert response.json['result']['templates']
    templates = response.json['result']['templates']
    assert len(templates) > 0
    default_template = templates[TEMPLATE_INDEX - 1]
    assert default_template['folder'] == TEMPLATE_ID


@pytest.mark.service
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_compare_manifests(svc_client_with_templates):
    """Check reading manifest template."""
    svc_client, headers, template_params = svc_client_with_templates

    response = svc_client.get(
        '/templates.read_manifest',
        query_string=template_params,
        headers=headers
    )

    assert response
    assert {'result'} == set(response.json.keys())
    assert response.json['result']['templates']

    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        manifest_file = fetch_template(TEMPLATE_URL, TEMPLATE_REF, temp_path)
        manifest = read_template_manifest(temp_path)

        assert manifest_file and manifest_file.exists()
        assert manifest

        templates_service = response.json['result']['templates']
        templates_local = manifest
        default_index = TEMPLATE_INDEX - 1
        assert templates_service[default_index
                                 ] == templates_local[default_index]



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of datasets."""
import os
from collections import defaultdict

import click

from ..echo import WARNING


def _dataset_pre_0_3(client):
    """Return paths of dataset metadata for pre 0.3.4."""
    project_is_pre_0_3 = int(client.project.version) < 2
    if project_is_pre_0_3:
        return (client.path / 'data').rglob(client.METADATA)
    return []


def check_dataset_metadata(client):
    """Check location of dataset metadata."""
    # Find pre 0.3.4 metadata files.
    old_metadata = list(_dataset_pre_0_3(client))

    if not old_metadata:
        return True, None

    problems = (
        WARNING + 'There are metadata files in the old location.'
        '\n  (use "renku migrate" to move them)\n\n\t' + '\n\t'.join(
            click.style(str(path.relative_to(client.path)), fg='yellow')
            for path in old_metadata
        ) + '\n'
    )

    return False, problems


def check_missing_files(client):
    """Find missing files listed in datasets."""
    missing = defaultdict(list)

    for dataset in client.datasets.values():
        for file_ in dataset.files:
            path = client.path / file_.path
            file_exists = (
                path.exists() or (file_.external and os.path.lexists(path))
            )
            if not file_exists:
                missing[dataset.name].append(file_.path)

    if not missing:
        return True, None

    problems = (WARNING + 'There are missing files in datasets.')

    for dataset, files in missing.items():
        problems += (
            '\n\t' + click.style(dataset, fg='yellow') + ':\n\t  ' +
            '\n\t  '.join(click.style(path, fg='red') for path in files)
        )

    return False, problems



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client utilities."""

import contextlib
import functools
import uuid

import click
import yaml

from renku.core.management import LocalClient

from ..management.config import RENKU_HOME
from ..management.repository import default_path
from .git import get_git_isolation


def _uuid_representer(dumper, data):
    """Add UUID serializer for YAML."""
    return dumper.represent_str(str(data))


yaml.add_representer(uuid.UUID, _uuid_representer)


def pass_local_client(
    method=None,
    clean=None,
    up_to_date=None,
    commit=None,
    commit_only=None,
    ignore_std_streams=True,
    commit_empty=True,
    raise_if_empty=False,
    lock=None,
):
    """Pass client from the current context to the decorated command."""
    if method is None:
        return functools.partial(
            pass_local_client,
            clean=clean,
            up_to_date=up_to_date,
            commit=commit,
            commit_only=commit_only,
            ignore_std_streams=ignore_std_streams,
            commit_empty=commit_empty,
            raise_if_empty=raise_if_empty,
            lock=lock,
        )

    def new_func(*args, **kwargs):
        ctx = click.get_current_context(silent=True)
        if ctx is None:
            client = LocalClient(
                path=default_path(),
                renku_home=RENKU_HOME,
                external_storage_requested=True,
            )
            ctx = click.Context(click.Command(method))
        else:
            client = ctx.ensure_object(LocalClient)

        stack = contextlib.ExitStack()

        # Handle --isolation option:
        if get_git_isolation():
            client = stack.enter_context(client.worktree())

        transaction = client.transaction(
            clean=clean,
            commit=commit,
            commit_empty=commit_empty,
            commit_message=kwargs.get('commit_message', None),
            commit_only=commit_only,
            ignore_std_streams=ignore_std_streams,
            raise_if_empty=raise_if_empty,
            up_to_date=up_to_date,
        )
        stack.enter_context(transaction)

        if lock or (lock is None and commit):
            stack.enter_context(client.lock)

        result = None
        if ctx:
            with stack:
                result = ctx.invoke(method, client, *args, **kwargs)

        return result

    return functools.update_wrapper(new_func, method)



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service exception tests for all endpoints."""
import pytest

from renku.service.config import INVALID_HEADERS_ERROR_CODE


@pytest.mark.service
def test_allowed_methods_exc(service_allowed_endpoint):
    """Check allowed methods for every endpoint."""
    methods, request, svc_client = service_allowed_endpoint

    method = request['allowed_method']
    if method == 'GET':  # if GET remove sister method HEAD
        methods.pop(method)
        methods.pop('HEAD')
    else:
        methods.pop(method)

    for method, fn in methods.items():
        response = fn(request['url'])
        assert 405 == response.status_code


@pytest.mark.service
def test_auth_headers_exc(service_allowed_endpoint):
    """Check correct headers for every endpoint."""
    methods, request, svc_client = service_allowed_endpoint

    method = request['allowed_method']
    if method == 'GET':  # if GET remove sister method HEAD
        client_method = methods.pop(method)
        methods.pop('HEAD')
    else:
        client_method = methods.pop(method)

    response = client_method(
        request['url'],
        headers=request['headers'],
    )

    assert 200 == response.status_code
    assert INVALID_HEADERS_ERROR_CODE == response.json['error']['code']

    err_message = 'user identification is incorrect or missing'
    assert err_message == response.json['error']['reason']
